{*******************************************************}
{                                                       }
{       Advanced SQL statement parser                   }
{       Classes for testing Select field parsing        }
{       Copyright (c) 2001 - 2003 AS Gaiasoft           }
{       Created by Gert Kello                           }
{                                                       }
{*******************************************************}

unit gaSQLSelectFieldTests;

interface

uses
  gaSQLParserTestersBase, gaSQLSelectFieldParsers;

type
  TgaSQLSelectFieldTester = class (TgaDefaultSQLParserTester)
  private
    FSelectField: TgaSQLSelectField;
  protected
    procedure SetUp; override;
    procedure TearDown; override;
    property SelectField: TgaSQLSelectField read FSelectField;
  published
    procedure PlainAsterix;
    procedure PlainFieldName;
    procedure PrefixedAsterix;
    procedure PrefixedAsterixWithAlias;
    procedure PrefixiedFieldName;
    procedure PrefixiedFieldNameWithAlias;
    procedure PrefixiedFieldNameWithAsAlias;
    procedure SelectExpression;
  end;
  
implementation

uses
  TestFrameWork, gaBasicSQLParser, gaAdvancedSQLParser,
  gaSQLParserHelperClasses, SysUtils;

{
*************************** TgaSQLSelectFieldTester ****************************
}
procedure TgaSQLSelectFieldTester.PlainAsterix;
begin
  AddStringToStatement('*', True);
  Check(SelectField.ParseComplete, 'Parse not complete after "*"');
  Check(not SelectField.IsInvalid, 'Part invalid after "*"');
  Check(SelectField.AsString = '*', 'AsString <> "*"');
  Check(SelectField.FieldName = '*', 'FieldName <> "*"');
  Check(SelectField.FieldPrefix = '', 'FieldPrefix <> ""');
  Check(SelectField.FieldAlias = '', 'FialdAlias <> ""');
end;

procedure TgaSQLSelectFieldTester.PlainFieldName;
begin
  AddStringToStatement('Tere', True);
  Check(not SelectField.ParseComplete, 'Parse complete after "Tere"');
  Check(not SelectField.IsInvalid, 'Part invalid after "Tere"');
  SelectField.ParseComplete := True;
  Check(SelectField.ParseComplete, 'Parse not complete after "SelectField.ParseComplete := True;"');
  Check(SelectField.AsString = 'Tere', 'AsString <> "Tere"');
  Check(SelectField.FieldName = 'Tere', 'FieldName <> "Tere"');
  Check(SelectField.FieldPrefix = '', 'FieldPrefix <> ""');
  Check(SelectField.FieldAlias = '', 'FialdAlias <> ""');
end;

procedure TgaSQLSelectFieldTester.PrefixedAsterix;
begin
  AddStringToStatement('Tere.Meie.*', True);
  Check(SelectField.ParseComplete, 'Parse not complete after "Tere.Meie.*"');
  Check(not SelectField.IsInvalid, 'Part invalid after "Tere.Meie.*"');
  Check(SelectField.AsString = 'Tere.Meie.*', 'AsString <> "Tere.Meie.*"');
  Check(SelectField.FieldName = '*', 'FieldName <> "*"');
  Check(SelectField.FieldPrefix = 'Tere.Meie.', 'FieldPrefix <> "Tere.Meie."');
  Check(SelectField.FieldAlias = '', 'FialdAlias <> ""');
end;

procedure TgaSQLSelectFieldTester.PrefixedAsterixWithAlias;
begin
  StartExpectingException(EgaSQLInvalidParseState);
  AddStringToStatement('Tere.Meie.* Tuut', True);
  StopExpectingException('Tere.Meie.*.Tuut should produce error ''Cant "ExcuteTokenAdded":....''');
end;

procedure TgaSQLSelectFieldTester.PrefixiedFieldName;
begin
  AddStringToStatement('Mina.Tere', True);
  Check(not SelectField.ParseComplete, 'Parse complete after "Mina.Tere"');
  Check(not SelectField.IsInvalid, 'Part invalid after "Mina.Tere"');
  SelectField.ParseComplete := True;
  Check(SelectField.ParseComplete, 'Parse not complete after "SelectField.ParseComplete := True;"');
  Check(SelectField.AsString = 'Mina.Tere', 'AsString <> "Mina.Tere"');
  Check(SelectField.FieldName = 'Tere', 'FieldName <> "Tere"');
  Check(SelectField.FieldPrefix = 'Mina.', 'FieldPrefix <> "Mina."');
  Check(SelectField.FieldAlias = '', 'FialdAlias <> ""');
end;

procedure TgaSQLSelectFieldTester.PrefixiedFieldNameWithAlias;
begin
  AddStringToStatement('Mina.Tere siin', True);
  Check(SelectField.ParseComplete, 'Parse not complete after "Mina.Tere siin"');
  Check(not SelectField.IsInvalid, 'Part invalid after "Mina.Tere siin"');
  Check(SelectField.AsString = 'Mina.Tere siin', 'AsString <> "Mina.Tere siin"');
  Check(SelectField.FieldName = 'Tere', 'FieldName <> "Tere"');
  Check(SelectField.FieldPrefix = 'Mina.', 'FieldPrefix <> "Mina."');
  Check(SelectField.FieldAlias = 'siin', 'FialdAlias <> "siin"');
end;

procedure TgaSQLSelectFieldTester.PrefixiedFieldNameWithAsAlias;
begin
  AddStringToStatement('Mina.Tere aS siin', True);
  Check(SelectField.ParseComplete, 'Parse not complete after "Mina.Tere siin"');
  Check(not SelectField.IsInvalid, 'Part invalid after "Mina.Tere siin"');
  Check(SelectField.AsString = 'Mina.Tere aS siin', 'AsString <> "Mina.Tere aS siin"');
  Check(SelectField.FieldName = 'Tere', 'FieldName <> "Tere"');
  Check(SelectField.FieldPrefix = 'Mina.', 'FieldPrefix <> "Mina."');
  Check(SelectField.FieldAlias = 'siin', 'FialdAlias <> "siin"');
end;

procedure TgaSQLSelectFieldTester.SelectExpression;
begin
  AddStringToStatement('(Mina + Sina)', True);
  Check(SelectField.FieldRef = nil, 'Expression, but Fieldref non-nil');
  Check(not SelectField.ValueExpression.IsInvalid, 'Expression considered to be invalid');
  Check(not SelectField.IsInvalid, 'Field reference considered to be invalid');
  Check(SelectField.FieldAlias = '', 'Where the heck did get alias?');
  Check(SelectField.ValueExpression.AsString = '(Mina + Sina)', 'Expression.AsString <> (Mina + Sina)');
end;

procedure TgaSQLSelectFieldTester.SetUp;
begin
  inherited;
  FSelectField := TgaSQLSelectField.Create(Statement);
  Statement.TestedStatementPart := SelectField;
end;

procedure TgaSQLSelectFieldTester.TearDown;
begin
  FSelectField.Free;
  inherited;
end;

end.
